@@ -15,7 +15,7 @@ use tc_transact::fs::{Dir, File, Hash};
 use tc_transact::{IntoView, Transaction, TxnId};
 use tc_value::Value;
 use tcgeneric::{
-    path_label, Class, Id, Instance, NativeClass, PathLabel, PathSegment, TCPathBuf, TCTryStream,
+    path_label, Class, Id, Instance, NativeClass, PathLabel, PathSegment, TCBoxTryStream, TCPathBuf,
 };
 
 use index::*;
@@ -31,58 +31,16 @@ mod index;
 mod schema;
 mod view;
 
+/// The key of a [`Table`] row.
+pub type Key = Vec<Value>;
+
+/// The values of a [`Table`] row.
+pub type Values = Vec<Value>;
+
 const PATH: PathLabel = path_label(&["state", "collection", "table"]);
 
-const ERR_DELETE: &str = "Deletion is not supported by instance of";
-const ERR_INSERT: &str = "Insertion is not supported by instance of";
-const ERR_SLICE: &str = "Slicing is not supported by instance of";
-const ERR_UPDATE: &str = "Update is not supported by instance of";
-
-/// Common [`Table`] methods.
-#[async_trait]
-pub trait TableInstance<F: File<Node>, D: Dir, Txn: Transaction<D>>:
-    Instance<Class = TableType> + Clone + Sized + Into<Table<F, D, Txn>>
-{
-    /// The type of `Table` returned by this instance's `order_by` method.
-    type OrderBy: TableInstance<F, D, Txn>;
-
-    /// The type of `Table` returned by this instance's `reversed` method.
-    type Reverse: TableInstance<F, D, Txn>;
-
-    /// The type of `Table` returned by this instance's `slice` method.
-    type Slice: TableInstance<F, D, Txn>;
-
-    /// Return the number of rows in this `Table`.
-    async fn count(self, txn_id: TxnId) -> TCResult<u64> {
-        let rows = self.rows(txn_id).await?;
-        rows.try_fold(0, |count, _| future::ready(Ok(count + 1)))
-            .await
-    }
-
-    /// Delete all rows in this `Table`.
-    async fn delete(&self, _txn_id: TxnId) -> TCResult<()> {
-        Err(TCError::bad_request(ERR_DELETE, self.class()))
-    }
-
-    /// Delete the given [`Row`] from this table, if present.
-    async fn delete_row(&self, _txn_id: TxnId, _row: Row) -> TCResult<()> {
-        Err(TCError::bad_request(ERR_DELETE, self.class()))
-    }
-
-    /// Group this `Table` by the given columns.
-    fn group_by(self, columns: Vec<Id>) -> TCResult<view::Aggregate<F, D, Txn, Self::OrderBy>> {
-        group_by(self, columns)
-    }
-
-    /// Construct and return a temporary index of the given columns.
-    async fn index(self, txn: Txn, columns: Option<Vec<Id>>) -> TCResult<index::ReadOnly<F, D, Txn>>
-    where
-        F: TryFrom<D::File, Error = TCError>,
-        D::FileClass: From<BTreeType>,
-    {
-        index::ReadOnly::copy_from(self, txn, columns).await
-    }
-
+/// Methods common to every [`Table`] view.
+pub trait TableInstance: Instance<Class = TableType> {
     /// Return the schema of this `Table`'s key.
     fn key(&self) -> &[Column];
 
@@ -91,65 +49,85 @@ pub trait TableInstance<F: File<Node>, D: Dir, Txn: Transaction<D>>:
 
     /// Return the schema of this `Table`.
     fn schema(&self) -> TableSchema;
+}
 
-    /// Limit the number of rows returned by `rows`.
-    fn limit(self, limit: u64) -> view::Limited<F, D, Txn> {
-        view::Limited::new(self.into(), limit)
-    }
+/// [`Table`] sort methods
+pub trait TableOrder: TableInstance {
+    /// The type of `Table` returned by this instance's `order_by` method.
+    type OrderBy: TableInstance;
 
-    /// Set the order returned by `rows`.
+    /// The type of `Table` returned by this instance's `reversed` method.
+    type Reverse: TableInstance;
+
+    /// Return an ordered view of this table.
     fn order_by(self, columns: Vec<Id>, reverse: bool) -> TCResult<Self::OrderBy>;
 
     /// Reverse the order returned by `rows`.
-    fn reversed(self) -> TCResult<Self::Reverse>;
-
-    /// Limit the columns returned by `rows`.
-    fn select(self, columns: Vec<Id>) -> TCResult<view::Selection<F, D, Txn, Self>> {
-        let selection = view::Selection::new(self, columns)?;
-        Ok(selection)
-    }
-
-    /// Limit the returned `rows` to the given [`Bounds`].
-    fn slice(self, _bounds: Bounds) -> TCResult<Self::Slice> {
-        Err(TCError::bad_request(ERR_SLICE, self.class()))
-    }
-
-    /// Return a stream of the rows in this `Table`.
-    async fn rows<'a>(self, txn_id: TxnId) -> TCResult<TCTryStream<'a, Vec<Value>>>;
-
-    /// Return an error if this table does not support the given [`Bounds`].
-    fn validate_bounds(&self, bounds: &Bounds) -> TCResult<()>;
+    fn reverse(self) -> TCResult<Self::Reverse>;
 
     /// Return an error if this table does not support ordering by the given columns.
     fn validate_order(&self, order: &[Id]) -> TCResult<()>;
+}
 
-    /// Update the values of the columns in this `Table` to match the given [`Row`].
-    async fn update(&self, _txn: &Txn, _value: Row) -> TCResult<()>
-    where
-        F: TryFrom<D::File, Error = TCError>,
-        D::FileClass: From<BTreeType>,
-    {
-        Err(TCError::bad_request(ERR_UPDATE, self.class()))
+/// Method to read a single row from a [`Table`]
+#[async_trait]
+pub trait TableRead: TableInstance {
+    async fn read(&self, txn_id: &TxnId, key: &Key) -> TCResult<Option<Vec<Value>>>;
+}
+
+/// Methods for slicing a [`Table`]
+pub trait TableSlice: TableStream {
+    /// The type of `Table` returned by this instance's `slice` method.
+    type Slice: TableInstance;
+
+    /// Limit the returned `rows` to the given [`Bounds`].
+    fn slice(self, _bounds: Bounds) -> TCResult<Self::Slice>;
+
+    /// Return an error if this table does not support the given [`Bounds`].
+    fn validate_bounds(&self, bounds: &Bounds) -> TCResult<()>;
+}
+
+/// [`Table`] read methods
+#[async_trait]
+pub trait TableStream: TableInstance + Sized {
+    type Limit: TableInstance;
+    type Selection: TableInstance;
+
+    /// Return the number of rows in this `Table`.
+    async fn count(self, txn_id: TxnId) -> TCResult<u64> {
+        let rows = self.rows(txn_id).await?;
+        rows.try_fold(0, |count, _| future::ready(Ok(count + 1)))
+            .await
     }
 
-    /// Update one row of this `Table`.
-    async fn update_row(&self, _txn_id: TxnId, _row: Row, _value: Row) -> TCResult<()> {
-        Err(TCError::bad_request(ERR_UPDATE, self.class()))
-    }
+    /// Limit the number of rows returned by `rows`.
+    fn limit(self, limit: u64) -> Self::Limit;
+
+    /// Limit the columns returned by `rows`.
+    fn select(self, columns: Vec<Id>) -> TCResult<Self::Selection>;
+
+    /// Return a stream of the rows in this `Table`.
+    async fn rows<'a>(self, txn_id: TxnId) -> TCResult<TCBoxTryStream<'a, Vec<Value>>>;
+}
+
+/// [`Table`] write methods
+#[async_trait]
+pub trait TableWrite: TableInstance {
+    /// Delete the given [`Row`] from this table, if present.
+    async fn delete(&self, txn_id: TxnId, key: Key) -> TCResult<()>;
+
+    /// Update one row of this table.
+    async fn update(&self, txn_id: TxnId, key: Key, values: Row) -> TCResult<()>;
 
     /// Insert or update the given row.
-    async fn upsert(&self, _txn_id: TxnId, _key: Vec<Value>, _value: Vec<Value>) -> TCResult<()> {
-        Err(TCError::bad_request(ERR_INSERT, self.class()))
-    }
+    async fn upsert(&self, txn_id: TxnId, key: Key, values: Values) -> TCResult<()>;
 }
 
 /// The [`Class`] of a [`Table`].
 #[derive(Clone, Copy, Hash, Eq, PartialEq)]
 pub enum TableType {
-    Index,
-    ReadOnly,
     Table,
-    Aggregate,
+    Index,
     IndexSlice,
     Limit,
     Merge,
@@ -184,10 +162,8 @@ impl Default for TableType {
 impl fmt::Display for TableType {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
-            Self::Index => write!(f, "type Index"),
-            Self::ReadOnly => write!(f, "type Index (read-only)"),
             Self::Table => write!(f, "type Table"),
-            Self::Aggregate => write!(f, "type Aggregate"),
+            Self::Index => write!(f, "type Index"),
             Self::IndexSlice => write!(f, "type Index slice"),
             Self::Limit => write!(f, "type Limit selection"),
             Self::Merge => write!(f, "type Merge selection"),
@@ -200,15 +176,13 @@ impl fmt::Display for TableType {
 /// An ordered collection of [`Row`]s which supports `BTree`-based indexing
 #[derive(Clone)]
 pub enum Table<F, D, Txn> {
-    Index(Index<F, D, Txn>),
-    ROIndex(ReadOnly<F, D, Txn>),
     Table(TableIndex<F, D, Txn>),
-    Aggregate(Box<Aggregate<F, D, Txn, Table<F, D, Txn>>>),
+    Index(Index<F, D, Txn>),
     IndexSlice(IndexSlice<F, D, Txn>),
     Limit(Box<Limited<F, D, Txn>>),
     Merge(Merged<F, D, Txn>),
     Selection(Box<Selection<F, D, Txn, Table<F, D, Txn>>>),
-    TableSlice(TableSlice<F, D, Txn>),
+    TableSlice(view::TableSlice<F, D, Txn>),
 }
 
 impl<F, D, Txn> Instance for Table<F, D, Txn>
@@ -219,10 +193,8 @@ where
 
     fn class(&self) -> Self::Class {
         match self {
-            Self::Index(_) => TableType::Index,
-            Self::ROIndex(_) => TableType::ReadOnly,
             Self::Table(_) => TableType::Table,
-            Self::Aggregate(_) => TableType::Aggregate,
+            Self::Index(_) => TableType::Index,
             Self::IndexSlice(_) => TableType::IndexSlice,
             Self::Limit(_) => TableType::Limit,
             Self::Merge(_) => TableType::Merge,
@@ -232,81 +204,14 @@ where
     }
 }
 
-#[async_trait]
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn> for Table<F, D, Txn>
+impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance for Table<F, D, Txn>
 where
     Self: Send + Sync,
 {
-    type OrderBy = Self;
-    type Reverse = Self;
-    type Slice = Self;
-
-    async fn count(self, txn_id: TxnId) -> TCResult<u64> {
-        match self {
-            Self::Index(index) => index.count(txn_id).await,
-            Self::ROIndex(index) => index.count(txn_id).await,
-            Self::Table(table) => table.count(txn_id).await,
-            Self::Aggregate(aggregate) => aggregate.count(txn_id).await,
-            Self::IndexSlice(slice) => slice.count(txn_id).await,
-            Self::Limit(limit) => limit.count(txn_id).await,
-            Self::Merge(merge) => merge.count(txn_id).await,
-            Self::Selection(selection) => selection.count(txn_id).await,
-            Self::TableSlice(slice) => slice.count(txn_id).await,
-        }
-    }
-
-    async fn delete(&self, txn_id: TxnId) -> TCResult<()> {
-        match self {
-            Self::Index(index) => index.delete(txn_id).await,
-            Self::ROIndex(index) => index.delete(txn_id).await,
-            Self::Table(table) => table.delete(txn_id).await,
-            Self::Aggregate(aggregate) => aggregate.delete(txn_id).await,
-            Self::IndexSlice(slice) => slice.delete(txn_id).await,
-            Self::Limit(limit) => limit.delete(txn_id).await,
-            Self::Merge(merge) => merge.delete(txn_id).await,
-            Self::Selection(selection) => selection.delete(txn_id).await,
-            Self::TableSlice(slice) => slice.delete(txn_id).await,
-        }
-    }
-
-    async fn delete_row(&self, txn_id: TxnId, row: Row) -> TCResult<()> {
-        match self {
-            Self::Index(index) => index.delete_row(txn_id, row).await,
-            Self::ROIndex(index) => index.delete_row(txn_id, row).await,
-            Self::Table(table) => table.delete_row(txn_id, row).await,
-            Self::Aggregate(aggregate) => aggregate.delete_row(txn_id, row).await,
-            Self::IndexSlice(slice) => slice.delete_row(txn_id, row).await,
-            Self::Limit(limit) => limit.delete_row(txn_id, row).await,
-            Self::Merge(merge) => merge.delete_row(txn_id, row).await,
-            Self::Selection(selection) => selection.delete_row(txn_id, row).await,
-            Self::TableSlice(slice) => slice.delete_row(txn_id, row).await,
-        }
-    }
-
-    async fn index(self, txn: Txn, columns: Option<Vec<Id>>) -> TCResult<index::ReadOnly<F, D, Txn>>
-    where
-        F: TryFrom<D::File, Error = TCError>,
-        D::FileClass: From<BTreeType>,
-    {
-        match self {
-            Self::Index(index) => index.index(txn, columns).await,
-            Self::ROIndex(index) => index.index(txn, columns).await,
-            Self::Table(table) => table.index(txn, columns).await,
-            Self::Aggregate(aggregate) => aggregate.index(txn, columns).await,
-            Self::IndexSlice(slice) => slice.index(txn, columns).await,
-            Self::Limit(limit) => limit.index(txn, columns).await,
-            Self::Merge(merge) => merge.index(txn, columns).await,
-            Self::Selection(selection) => selection.index(txn, columns).await,
-            Self::TableSlice(slice) => slice.index(txn, columns).await,
-        }
-    }
-
     fn key(&self) -> &[Column] {
         match self {
-            Self::Index(index) => index.key(),
-            Self::ROIndex(index) => index.key(),
             Self::Table(table) => table.key(),
-            Self::Aggregate(aggregate) => aggregate.key(),
+            Self::Index(index) => index.key(),
             Self::IndexSlice(slice) => slice.key(),
             Self::Limit(limit) => limit.key(),
             Self::Merge(merge) => merge.key(),
@@ -317,10 +222,8 @@ where
 
     fn values(&self) -> &[Column] {
         match self {
-            Self::Index(index) => index.values(),
-            Self::ROIndex(index) => index.values(),
             Self::Table(table) => table.values(),
-            Self::Aggregate(aggregate) => aggregate.values(),
+            Self::Index(slice) => slice.values(),
             Self::IndexSlice(slice) => slice.values(),
             Self::Limit(limit) => limit.values(),
             Self::Merge(merge) => merge.values(),
@@ -331,10 +234,8 @@ where
 
     fn schema(&self) -> TableSchema {
         match self {
-            Self::Index(index) => TableInstance::schema(index),
-            Self::ROIndex(index) => index.schema(),
             Self::Table(table) => table.schema(),
-            Self::Aggregate(aggregate) => aggregate.schema(),
+            Self::Index(slice) => TableInstance::schema(slice),
             Self::IndexSlice(slice) => TableInstance::schema(slice),
             Self::Limit(limit) => limit.schema(),
             Self::Merge(merge) => merge.schema(),
@@ -342,69 +243,125 @@ where
             Self::TableSlice(slice) => slice.schema(),
         }
     }
+}
 
-    fn limit(self, limit: u64) -> view::Limited<F, D, Txn> {
-        match self {
-            Self::Index(index) => index.limit(limit),
-            Self::ROIndex(index) => index.limit(limit),
-            Self::Table(table) => table.limit(limit),
-            Self::Aggregate(aggregate) => aggregate.limit(limit),
-            Self::IndexSlice(slice) => slice.limit(limit),
-            Self::Limit(limited) => limited.limit(limit),
-            Self::Merge(merge) => merge.limit(limit),
-            Self::Selection(selection) => selection.limit(limit),
-            Self::TableSlice(slice) => slice.limit(limit),
-        }
-    }
+#[async_trait]
+impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableOrder for Table<F, D, Txn>
+where
+    Self: Send + Sync,
+{
+    type OrderBy = Self;
+    type Reverse = Self;
 
     fn order_by(self, order: Vec<Id>, reverse: bool) -> TCResult<Self::OrderBy> {
         match self {
-            Self::Index(index) => index.order_by(order, reverse).map(Self::from),
-            Self::ROIndex(index) => index.order_by(order, reverse).map(Self::from),
             Self::Table(table) => table.order_by(order, reverse).map(Self::from),
-            Self::Aggregate(aggregate) => aggregate.order_by(order, reverse).map(Self::from),
+            Self::Index(index) => index.order_by(order, reverse).map(Self::from),
             Self::IndexSlice(slice) => slice.order_by(order, reverse).map(Self::from),
-            Self::Limit(limited) => limited.order_by(order, reverse).map(Self::from),
             Self::Merge(merge) => merge.order_by(order, reverse).map(Self::from),
             Self::Selection(selection) => selection.order_by(order, reverse).map(Self::from),
             Self::TableSlice(slice) => slice.order_by(order, reverse).map(Self::from),
+            other => Err(TCError::unsupported(format!(
+                "instance of {} does not support ordering",
+                other.class()
+            ))),
         }
     }
 
-    fn reversed(self) -> TCResult<Self::Reverse> {
+    fn reverse(self) -> TCResult<Self::Reverse> {
         match self {
-            Self::Index(index) => index.reversed().map(Self::from),
-            Self::ROIndex(index) => index.reversed().map(Self::from),
-            Self::Table(table) => table.reversed().map(Self::from),
-            Self::Aggregate(aggregate) => aggregate.reversed().map(Self::from),
-            Self::IndexSlice(slice) => slice.reversed().map(Self::from),
-            Self::Limit(limited) => limited.reversed().map(Self::from),
-            Self::Merge(merge) => merge.reversed().map(Self::from),
-            Self::Selection(selection) => selection.reversed().map(Self::from),
-            Self::TableSlice(slice) => slice.reversed().map(Self::from),
+            Self::Table(table) => table.reverse().map(Self::from),
+            Self::Index(index) => index.reverse().map(Self::from),
+            Self::IndexSlice(slice) => slice.reverse().map(Self::from),
+            Self::Merge(merge) => merge.reverse().map(Self::from),
+            Self::Selection(selection) => selection.reverse().map(Self::from),
+            Self::TableSlice(slice) => slice.reverse().map(Self::from),
+            other => Err(TCError::unsupported(format!(
+                "instance of {} does not support ordering",
+                other.class()
+            ))),
         }
     }
 
-    fn slice(self, bounds: Bounds) -> TCResult<Table<F, D, Txn>> {
+    fn validate_order(&self, order: &[Id]) -> TCResult<()> {
         match self {
-            Self::Index(index) => index.slice(bounds).map(Self::from),
-            Self::ROIndex(index) => index.slice(bounds).map(Self::from),
-            Self::Table(table) => table.slice(bounds).map(Self::from),
-            Self::Aggregate(aggregate) => aggregate.slice(bounds).map(Self::from),
-            Self::IndexSlice(slice) => slice.slice(bounds).map(Self::from),
-            Self::Limit(limited) => limited.slice(bounds).map(Self::from),
-            Self::Merge(merge) => merge.slice(bounds).map(Self::from),
-            Self::Selection(selection) => selection.slice(bounds).map(Self::from),
-            Self::TableSlice(slice) => slice.slice(bounds).map(Self::from),
+            Self::Table(table) => table.validate_order(order),
+            Self::Index(index) => index.validate_order(order),
+            Self::IndexSlice(slice) => slice.validate_order(order),
+            Self::Merge(merge) => merge.validate_order(order),
+            Self::Selection(selection) => selection.validate_order(order),
+            Self::TableSlice(slice) => slice.validate_order(order),
+            other => Err(TCError::unsupported(format!(
+                "instance of {} does not support ordering",
+                other.class()
+            ))),
+        }
+    }
+}
+#[async_trait]
+impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableRead for Table<F, D, Txn>
+where
+    Self: Send + Sync,
+{
+    async fn read(&self, txn_id: &TxnId, key: &Key) -> TCResult<Option<Vec<Value>>> {
+        match self {
+            Self::Table(table) => table.read(txn_id, key).await,
+            other => Err(TCError::unsupported(format!(
+                "{} does not support GET by key",
+                other
+            ))),
+        }
+    }
+}
+
+#[async_trait]
+impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableStream for Table<F, D, Txn>
+where
+    Self: Send + Sync,
+{
+    type Limit = Self;
+    type Selection = Self;
+
+    async fn count(self, txn_id: TxnId) -> TCResult<u64> {
+        match self {
+            Self::Table(table) => table.count(txn_id).await,
+            Self::Index(index) => index.count(txn_id).await,
+            Self::IndexSlice(slice) => slice.count(txn_id).await,
+            Self::Limit(limit) => limit.count(txn_id).await,
+            Self::Merge(merge) => merge.count(txn_id).await,
+            Self::Selection(selection) => selection.count(txn_id).await,
+            Self::TableSlice(slice) => slice.count(txn_id).await,
         }
     }
 
-    async fn rows<'a>(self, txn_id: TxnId) -> TCResult<TCTryStream<'a, Vec<Value>>> {
+    fn limit(self, limit: u64) -> <Self as TableStream>::Limit {
+        match self {
+            Self::Table(table) => table.limit(limit).into(),
+            Self::Index(index) => index.limit(limit).into(),
+            Self::IndexSlice(slice) => slice.limit(limit).into(),
+            Self::Limit(limited) => limited.limit(limit).into(),
+            Self::Merge(merge) => merge.limit(limit).into(),
+            Self::Selection(selection) => selection.limit(limit).into(),
+            Self::TableSlice(slice) => slice.limit(limit).into(),
+        }
+    }
+
+    fn select(self, columns: Vec<Id>) -> TCResult<<Self as TableStream>::Selection> {
+        match self {
+            Self::Table(table) => table.select(columns).map(Self::from),
+            Self::Index(index) => index.select(columns).map(Self::from),
+            Self::IndexSlice(slice) => slice.select(columns).map(Self::from),
+            Self::Limit(limited) => limited.select(columns).map(Self::from),
+            Self::Merge(merge) => merge.select(columns).map(Self::from),
+            Self::Selection(selection) => selection.select(columns).map(Self::from),
+            Self::TableSlice(slice) => slice.select(columns).map(Self::from),
+        }
+    }
+
+    async fn rows<'a>(self, txn_id: TxnId) -> TCResult<TCBoxTryStream<'a, Vec<Value>>> {
         match self {
-            Self::Index(index) => index.rows(txn_id).await,
-            Self::ROIndex(index) => index.rows(txn_id).await,
             Self::Table(table) => table.rows(txn_id).await,
-            Self::Aggregate(aggregate) => aggregate.rows(txn_id).await,
+            Self::Index(index) => index.rows(txn_id).await,
             Self::IndexSlice(slice) => slice.rows(txn_id).await,
             Self::Limit(limited) => limited.rows(txn_id).await,
             Self::Merge(merge) => merge.rows(txn_id).await,
@@ -412,78 +369,75 @@ where
             Self::TableSlice(slice) => slice.rows(txn_id).await,
         }
     }
+}
+
+#[async_trait]
+impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableSlice for Table<F, D, Txn>
+where
+    Self: Send + Sync,
+{
+    type Slice = Self;
+
+    fn slice(self, bounds: Bounds) -> TCResult<Table<F, D, Txn>> {
+        match self {
+            Self::Table(table) => table.slice(bounds).map(Self::from),
+            Self::Merge(merge) => merge.slice(bounds).map(Self::from),
+            Self::TableSlice(slice) => slice.slice(bounds).map(Self::from),
+            other => Err(TCError::unsupported(format!(
+                "instance of {} does not support slicing",
+                other.class()
+            ))),
+        }
+    }
 
     fn validate_bounds(&self, bounds: &Bounds) -> TCResult<()> {
         match self {
-            Self::Index(index) => index.validate_bounds(bounds),
-            Self::ROIndex(index) => index.validate_bounds(bounds),
             Self::Table(table) => table.validate_bounds(bounds),
-            Self::Aggregate(aggregate) => aggregate.validate_bounds(bounds),
-            Self::IndexSlice(slice) => slice.validate_bounds(bounds),
-            Self::Limit(limited) => limited.validate_bounds(bounds),
             Self::Merge(merge) => merge.validate_bounds(bounds),
-            Self::Selection(selection) => selection.validate_bounds(bounds),
             Self::TableSlice(slice) => slice.validate_bounds(bounds),
+            other => Err(TCError::unsupported(format!(
+                "instance of {} does not support slicing",
+                other.class()
+            ))),
+        }
+    }
+}
+
+#[async_trait]
+impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableWrite for Table<F, D, Txn>
+where
+    Self: Send + Sync,
+{
+    async fn delete(&self, txn_id: TxnId, key: Key) -> TCResult<()> {
+        if let Self::Table(table) = self {
+            table.delete(txn_id, key).await
+        } else {
+            Err(TCError::unsupported(format!(
+                "instance of {} does not support delete",
+                self.class()
+            )))
         }
     }
 
-    fn validate_order(&self, order: &[Id]) -> TCResult<()> {
-        match self {
-            Self::Index(index) => index.validate_order(order),
-            Self::ROIndex(index) => index.validate_order(order),
-            Self::Table(table) => table.validate_order(order),
-            Self::Aggregate(aggregate) => aggregate.validate_order(order),
-            Self::IndexSlice(slice) => slice.validate_order(order),
-            Self::Limit(limited) => limited.validate_order(order),
-            Self::Merge(merge) => merge.validate_order(order),
-            Self::Selection(selection) => selection.validate_order(order),
-            Self::TableSlice(slice) => slice.validate_order(order),
+    async fn update(&self, txn_id: TxnId, key: Key, values: Row) -> TCResult<()> {
+        if let Self::Table(table) = self {
+            table.update(txn_id, key, values).await
+        } else {
+            Err(TCError::unsupported(format!(
+                "instance of {} does not support delete",
+                self.class()
+            )))
         }
     }
 
-    async fn update(&self, txn: &Txn, value: Row) -> TCResult<()>
-    where
-        F: TryFrom<D::File, Error = TCError>,
-        D::FileClass: From<BTreeType>,
-    {
-        match self {
-            Self::Index(index) => index.update(txn, value).await,
-            Self::ROIndex(index) => index.update(txn, value).await,
-            Self::Table(table) => table.update(txn, value).await,
-            Self::Aggregate(aggregate) => aggregate.update(txn, value).await,
-            Self::IndexSlice(slice) => slice.update(txn, value).await,
-            Self::Limit(limited) => limited.update(txn, value).await,
-            Self::Merge(merge) => merge.update(txn, value).await,
-            Self::Selection(selection) => selection.update(txn, value).await,
-            Self::TableSlice(slice) => slice.update(txn, value).await,
-        }
-    }
-
-    async fn update_row(&self, txn_id: TxnId, row: Row, value: Row) -> TCResult<()> {
-        match self {
-            Self::Index(index) => index.update_row(txn_id, row, value).await,
-            Self::ROIndex(index) => index.update_row(txn_id, row, value).await,
-            Self::Table(table) => table.update_row(txn_id, row, value).await,
-            Self::Aggregate(aggregate) => aggregate.update_row(txn_id, row, value).await,
-            Self::IndexSlice(slice) => slice.update_row(txn_id, row, value).await,
-            Self::Limit(limited) => limited.update_row(txn_id, row, value).await,
-            Self::Merge(merge) => merge.update_row(txn_id, row, value).await,
-            Self::Selection(selection) => selection.update_row(txn_id, row, value).await,
-            Self::TableSlice(slice) => slice.update_row(txn_id, row, value).await,
-        }
-    }
-
-    async fn upsert(&self, txn_id: TxnId, key: Vec<Value>, values: Vec<Value>) -> TCResult<()> {
-        match self {
-            Self::Index(index) => index.upsert(txn_id, key, values).await,
-            Self::ROIndex(index) => index.upsert(txn_id, key, values).await,
-            Self::Table(table) => table.upsert(txn_id, key, values).await,
-            Self::Aggregate(aggregate) => aggregate.upsert(txn_id, key, values).await,
-            Self::IndexSlice(slice) => slice.upsert(txn_id, key, values).await,
-            Self::Limit(limited) => limited.upsert(txn_id, key, values).await,
-            Self::Merge(merge) => merge.upsert(txn_id, key, values).await,
-            Self::Selection(selection) => selection.upsert(txn_id, key, values).await,
-            Self::TableSlice(slice) => slice.upsert(txn_id, key, values).await,
+    async fn upsert(&self, txn_id: TxnId, key: Key, values: Values) -> TCResult<()> {
+        if let Self::Table(table) = self {
+            table.upsert(txn_id, key, values).await
+        } else {
+            Err(TCError::unsupported(format!(
+                "instance of {} does not support delete",
+                self.class()
+            )))
         }
     }
 }
@@ -493,7 +447,7 @@ impl<'en, F: File<Node>, D: Dir, Txn: Transaction<D>> Hash<'en, D> for Table<F,
     type Item = Vec<Value>;
     type Txn = Txn;
 
-    async fn hashable(&'en self, txn: &'en Txn) -> TCResult<TCTryStream<'en, Self::Item>> {
+    async fn hashable(&'en self, txn: &'en Txn) -> TCResult<TCBoxTryStream<'en, Self::Item>> {
         self.clone().rows(*txn.id()).await
     }
 }
@@ -571,6 +525,7 @@ where
             .next_element(())
             .await?
             .ok_or_else(|| de::Error::invalid_length(0, "a Table schema"))?;
+
         let table = TableIndex::create(self.txn.context(), schema, *self.txn.id())
             .map_err(de::Error::custom)
             .await?;
@@ -604,7 +559,10 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> de::Visitor for RowVisitor<F, D
 
         while let Some(row) = seq.next_element(()).await? {
             let row = schema.row_from_values(row).map_err(de::Error::custom)?;
-            let (key, values) = schema.key_values_from_row(row).map_err(de::Error::custom)?;
+            let (key, values) = schema
+                .key_values_from_row(row, true)
+                .map_err(de::Error::custom)?;
+
             self.table
                 .upsert(self.txn_id, key, values)
                 .map_err(de::Error::custom)
@@ -631,7 +589,7 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> de::FromStream for RowVisitor<F
 /// A view of a [`Table`] within a single [`Transaction`], used for serialization.
 pub struct TableView<'en> {
     schema: TableSchema,
-    rows: TCTryStream<'en, Vec<Value>>,
+    rows: TCBoxTryStream<'en, Vec<Value>>,
 }
 
 impl<'en> en::IntoStream<'en> for TableView<'en> {
